#pragma once

#include "syringe.h"
#include "motor.h"
#include <ArduinoJson.h>
#include "math.h"
#include <interrupts.h> // InterruptLock
using InterruptLock = esp8266::InterruptLock;

#include "fs.h"

#include "common.h"


class SyringeFilled : public Motor
{
    /***
     * Role of this class :
     * -Store information about liquid
     * -Get informations about movement required
     * -Manage movements
    
    ***/

    private :

    //SyringeFilled
    float _exchange_throughput_uL_per_min;
    float _exchange_volume_mL;
    float _remaining_volume_mL;
    bool _push;  //false means pull
    bool _throughput; //false means duration
    bool _update_remaining_volume;

    //Syringe
    String _name_syringe; //Not sure how to manage it

    //Syringe_Pump
    bool _clockwise_equals_push;
    float _syringe_pump_length_mm;
    float _lead_screw_pitch_mm;
    int _microstep_configuration;
    int _motor_steps;
    bool _motor_release;

    //Limit_Switch
    bool _emergency;

    //JSON DOCUMENT
    StaticJsonDocument<400> syringe_filled_json;  //400 = RAM allocated to this document


    public :
    
    //CONSTRUCTORS
    SyringeFilled(MotorHardware_t& stepper); 

    //SET METHODS
    void set_exchange_throughput_uL_per_min(float exchange_throughput_uL_per_min);
    void set_exchange_volume_mL(float exchange_volume_mL);
    void set_remaining_volume_mL(float remaining_volume_mL);
    void set_push(bool push);
    void set_throughput(bool throughput);
    void set_update_remaining_volume(bool update_remaining_volume);
    void set_name_syringe(String name_syringe);
    void set_clockwise_equals_push(bool clockwise_equals_push);
    void set_lead_screw_pitch_mm(float lead_screw_pitch_mm);
    void set_syringe_pump_length_mm(float syringe_pump_length_mm);
    void set_microstep_configuration(int microstep_configuration);
    void set_motor_steps(int motor_steps);
    void set_motor_release(int motor_release);
    void set_emergency(bool emergency);


    //GET METHODS
    float get_exchange_throughput_uL_per_min();
    float get_exchange_volume_mL();
    float get_remaining_volume_mL();
    bool get_push();
    bool get_throughput();
    bool get_update_remaining_volume();
    String get_syringe_name();
    bool get_clockwise_equals_push();
    float get_syringe_pump_length_mm();
    float get_lead_screw_pitch_mm();
    int get_microstep_configuration();
    int get_motor_steps();
    bool get_motor_release();
    bool get_emergency();
    const StaticJsonDocument<200>& get_syringe_filled_data();





    //CONVERSIONS
    float volume_mL_to_distance_mm(float volume_mL);
    float distance_mm_to_volume_mL(float distance_mm); //CLI

    //MOVEMENTS
    bool is_running ();
    void start_exchange();
    void go_to_limit_switch(); //CLI

    //LIMIT SWITCH
    void manage_emergency (bool pressed, bool stepperMoving);
    void run_from_emergency(); //CLI

    //CONFIGURATIONS
    bool check_configuration(); //CLI
    void show_configuration(); //CLI
    
    //JSON
    void load_json ();
    void save_json ();




};

extern SyringeFilled syringe_filled;

